/* -*-c-*-
 *
 * Definitions for Catacomb bindings
 *
 * (c) 2004 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of the Python interface to Catacomb.
 *
 * Catacomb/Python is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Catacomb/Python is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Catacomb/Python; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef CATACOMB_PYTHON_H
#define CATACOMB_PYTHON_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#define PY_SSIZE_T_CLEAN

#include <Python.h>
#include <longintrepr.h>
#include <structmember.h>

#include <mLib/darray.h>
#include <mLib/dstr.h>
#include <mLib/macros.h>
#include <mLib/quis.h>
#include <mLib/unihash.h>

#include <catacomb/buf.h>
#include <catacomb/ct.h>

#include <catacomb/grand.h>
#include <catacomb/rand.h>
#include <catacomb/noise.h>
#include <catacomb/bbs.h>
#include <catacomb/mprand.h>
#include <catacomb/lcrand.h>
#include <catacomb/fibrand.h>
#include <catacomb/dsarand.h>
#include <catacomb/sslprf.h>
#include <catacomb/tlsprf.h>
#include <catacomb/blkc.h>

#include <catacomb/gcipher.h>
#include <catacomb/gaead.h>
#include <catacomb/ghash.h>
#include <catacomb/gmac.h>
#include <catacomb/md5.h>
#include <catacomb/md5-hmac.h>
#include <catacomb/poly1305.h>
#include <catacomb/sha.h>
#include <catacomb/sha-mgf.h>
#include <catacomb/sha-hmac.h>
#include <catacomb/keccak1600.h>
#include <catacomb/sha3.h>

#include <catacomb/mp.h>
#include <catacomb/mpint.h>
#include <catacomb/mpmul.h>
#include <catacomb/mpcrt.h>
#include <catacomb/mpmont.h>
#include <catacomb/mpbarrett.h>
#include <catacomb/mpreduce.h>
#include <catacomb/mp-fibonacci.h>

#include <catacomb/pgen.h>
#include <catacomb/pfilt.h>
#include <catacomb/strongprime.h>
#include <catacomb/limlee.h>
#include <catacomb/dh.h>
#include <catacomb/ptab.h>
#include <catacomb/bintab.h>
#include <catacomb/dsa.h>
#include <catacomb/x25519.h>
#include <catacomb/x448.h>
#include <catacomb/ed25519.h>
#include <catacomb/ed448.h>

#include <catacomb/gf.h>
#include <catacomb/gfreduce.h>
#include <catacomb/gfn.h>

#include <catacomb/field.h>
#include <catacomb/field-guts.h>

#include <catacomb/ec.h>
#include <catacomb/ec-raw.h>
#include <catacomb/ectab.h>

#include <catacomb/group.h>
#include <catacomb/group-guts.h>

#include <catacomb/gdsa.h>
#include <catacomb/gkcdsa.h>
#include <catacomb/rsa.h>

#include <catacomb/key.h>
#include <catacomb/passphrase.h>
#include <catacomb/pixie.h>

#include <catacomb/share.h>
#include <catacomb/gfshare.h>

/*----- Other preliminaries -----------------------------------------------*/

#define GOBBLE_SEMI extern int notexist
#if defined(__GNUC__) && defined(__ELF__)
#  define PRIVATE_SYMBOLS _Pragma("GCC visibility push(hidden)") GOBBLE_SEMI
#  define PUBLIC_SYMBOLS _Pragma("GCC visibility pop") GOBBLE_SEMI
#  define EXPORT __attribute__((__visibility__("default")))
#else
#  define PRIVATE_SYMBOLS GOBBLE_SEMI
#  define PUBLIC_SYMBOLS GOBBLE_SEMI
#  define EXPORT
#endif

PRIVATE_SYMBOLS;

/*----- Utility macros ----------------------------------------------------*/

#define RETURN_OBJ(obj) do { Py_INCREF(obj); return (obj); } while (0)
#define RETURN_NONE RETURN_OBJ(Py_None)
#define RETURN_NOTIMPL RETURN_OBJ(Py_NotImplemented)
#define RETURN_TRUE RETURN_OBJ(Py_True)
#define RETURN_FALSE RETURN_OBJ(Py_False)
#define RETURN_ME RETURN_OBJ(me)

#define EXCERR(exc, str) do {						\
  PyErr_SetString(exc, str);						\
  goto end;								\
} while (0)
#define VALERR(str) EXCERR(PyExc_ValueError, str)
#define OVFERR(str) EXCERR(PyExc_OverflowError, str)
#define TYERR(str) EXCERR(PyExc_TypeError, str)
#define IXERR(str) EXCERR(PyExc_IndexError, str)
#define ZDIVERR(str) EXCERR(PyExc_ZeroDivisionError, str)
#define SYSERR(str) EXCERR(PyExc_SystemError, str)
#define NIERR(str) EXCERR(PyExc_NotImplementedError, str)
#define INDEXERR(idx) do {						\
  PyErr_SetObject(PyExc_KeyError, idx);					\
  goto end;								\
} while (0)
#define OSERR(name) do {						\
  PyErr_SetFromErrnoWithFilename(PyExc_OSError, name);			\
  goto end;								\
} while (0)
#define PGENERR(exc) do { pgenerr(exc); goto end; } while (0)

#define CONVFUNC(ty, cty, ext)						\
  int conv##ty(PyObject *o, void *p)					\
  {									\
    if (!PyObject_TypeCheck(o, ty##_pytype))				\
      TYERR("wanted a " #ty);						\
    *(cty *)p = ext(o);							\
    return (1);								\
  end:									\
    return (0);								\
  }

#define root_pytype 0
#define type_pytype &PyType_Type
#define INITTYPE_META(ty, base, meta) do {				\
  ty##_pytype_skel.tp_base = base##_pytype;				\
  ty##_pytype = inittype(&ty##_pytype_skel, meta##_pytype);		\
} while (0)
#define INITTYPE(ty, base) INITTYPE_META(ty, base, type)

extern PyObject *home_module;
extern PyObject *modname;

#define INSERT(name, ob) do {	\
  PyObject *_o = (PyObject *)(ob);					\
  Py_INCREF(_o);							\
  PyModule_AddObject(mod, name, _o);					\
} while (0)

#define INSEXC(name, var, base, meth)					\
  INSERT(name, var = mkexc(mod, base, name, meth))

#define METH(func, doc)							\
  { #func, METHNAME(func), METH_VARARGS, doc },
#define KWMETH(func, doc)						\
  { #func, (PyCFunction)METHNAME(func),					\
    METH_VARARGS | METH_KEYWORDS, doc },

#define GET(func, doc)							\
  { #func, GETSETNAME(get, func), 0, doc },
#define GETSET(func, doc)						\
  { #func, GETSETNAME(get, func), GETSETNAME(set, func), doc },

#define MEMBER(name, ty, f, doc)					\
  { #name, ty, offsetof(MEMBERSTRUCT, name), f, doc },

#define MODULES(_)							\
  _(util)								\
  _(bytestring) _(buffer)						\
  _(rand) _(algorithms) _(pubkey) _(pgen)				\
  _(mp) _(field) _(ec) _(group)						\
  _(passphrase) _(share) _(key)
#define DOMODINIT(m) m##_pyinit();
#define DOMODINSERT(m) m##_pyinsert(mod);
#define INIT_MODULES do { MODULES(DOMODINIT) } while (0)
#define INSERT_MODULES do { MODULES(DOMODINSERT) } while (0)

#define DO(m)								\
  extern void m##_pyinit(void);						\
  extern void m##_pyinsert(PyObject *);
MODULES(DO)
#undef DO

#define FREEOBJ(obj)							\
  (((PyObject *)(obj))->ob_type->tp_free((PyObject *)(obj)))

#define GEN(func, base)							\
  static PyObject *func(void)						\
  {									\
    PyObject *d = PyDict_New();						\
    PyObject *o;							\
    int i;								\
									\
    for (i = 0; g##base##tab[i]; i++) {					\
      o = gc##base##_pywrap((/*unconst*/ gc##base *)g##base##tab[i]);	\
      PyDict_SetItemString(d,						\
			   (/*unconst*/ char *)g##base##tab[i]->name,	\
			   o);						\
      Py_DECREF(o);							\
    }									\
    return (d);								\
  }

#define KWLIST (/*unconst*/ char **)kwlist

struct nameval { const char *name; unsigned f; unsigned long value; };
#define CF_SIGNED 1u
extern void setconstants(PyObject *, const struct nameval *);

extern PyObject *mexp_common(PyObject *, PyObject *, size_t,
			     PyObject *(*id)(PyObject *),
			     int (*fill)(void *, PyObject *,
					 PyObject *, PyObject *),
			     PyObject *(*exp)(PyObject *, void *, int),
			     void (*drop)(void *));

extern int convulong(PyObject *, void *);
#define DECL_CONVU_(n) extern int convu##n(PyObject *, void *);
DOUINTSZ(DECL_CONVU_)
extern int convmpw(PyObject *, void *);
extern int convuint(PyObject *, void *);
extern int convk64(PyObject *, void *);
extern int convszt(PyObject *, void *);
extern int convbool(PyObject *, void *);
extern PyObject *abstract_pynew(PyTypeObject *, PyObject *, PyObject *);
extern PyObject *getbool(int);
extern PyObject *getulong(unsigned long);
extern PyObject *getk64(kludge64);
extern void *newtype(PyTypeObject *, const PyTypeObject *, const char *);

struct excinfo { PyObject *ty, *val, *tb; };
#define EXCINFO_INIT { 0, 0, 0 }

extern PyObject *mkexc(PyObject *, PyObject *, const char *, PyMethodDef *);
#define INIT_EXCINFO(exc) do {						\
  struct excinfo *_exc = (exc); _exc->ty = _exc->val = _exc->tb = 0;	\
} while (0)
#define RELEASE_EXCINFO(exc) do {					\
  struct excinfo *_exc = (exc);						\
  Py_XDECREF(_exc->ty);	 _exc->ty  = 0;					\
  Py_XDECREF(_exc->val); _exc->val = 0;					\
  Py_XDECREF(_exc->tb);	 _exc->tb  = 0;					\
} while (0)
#define STASH_EXCINFO(exc) do {						\
  struct excinfo *_exc = (exc);						\
  PyErr_Fetch(&_exc->ty, &_exc->val, &_exc->tb);			\
  PyErr_NormalizeException(&_exc->ty, &_exc->val, &_exc->tb);		\
} while (0)
#define RESTORE_EXCINFO(exc) do {					\
  struct excinfo *_exc = (exc);						\
  PyErr_Restore(_exc->ty, _exc->val, _exc->tb);				\
  _exc->ty = _exc->val = _exc->tb = 0;					\
} while (0)
extern void report_lost_exception(struct excinfo *, const char *, ...);
extern void report_lost_exception_v(struct excinfo *, const char *, va_list);
extern void stash_exception(struct excinfo *, const char *, ...);
extern void restore_exception(struct excinfo *, const char *, ...);

extern void typeready(PyTypeObject *);
extern PyTypeObject *inittype(PyTypeObject *, PyTypeObject *);
extern void addmethods(const PyMethodDef *);
extern PyMethodDef *donemethods(void);

/*----- Mapping methods ---------------------------------------------------*/

#define GMAP_METH(func, doc) { #func, gmapmeth_##func, METH_VARARGS, doc },
#define GMAP_KWMETH(func, doc)						\
  { #func, (PyCFunction)gmapmeth_##func, METH_VARARGS|METH_KEYWORDS, doc },
#define GMAP_METHDECL(func, doc)					\
  extern PyObject *gmapmeth_##func(PyObject *, PyObject *);
#define GMAP_KWMETHDECL(func, doc)					\
  extern PyObject *gmapmeth_##func(PyObject *, PyObject *, PyObject *);

#define GMAP_DOROMETHODS(METH, KWMETH)					\
  METH	(has_key,	"D.has_key(KEY) -> BOOL")			\
  METH	(keys,		"D.keys() -> LIST")				\
  METH	(values,	"D.values() -> LIST")				\
  METH	(items,		"D.items() -> LIST")				\
  METH	(iterkeys,	"D.iterkeys() -> ITER")				\
  METH	(itervalues,	"D.itervalues() -> ITER")			\
  METH	(iteritems,	"D.iteritems() -> ITER")			\
  KWMETH(get,		"D.get(KEY, [default = None]) -> VALUE")	\

#define GMAP_DOMETHODS(METH, KWMETH)					\
  GMAP_DOROMETHODS(METH, KWMETH)					\
  METH	(clear,		"D.clear()")					\
  KWMETH(setdefault,	"D.setdefault(K, [default = None]) -> VALUE")	\
  KWMETH(pop,		"D.pop(KEY, [default = <error>]) -> VALUE")	\
  METH	(popitem,	"D.popitem() -> (KEY, VALUE)")			\
  METH	(update,	"D.update(MAP)")

GMAP_DOMETHODS(GMAP_METHDECL, GMAP_KWMETHDECL)
#define GMAP_ROMETHODS GMAP_DOROMETHODS(GMAP_METH, GMAP_KWMETH)
#define GMAP_METHODS GMAP_DOMETHODS(GMAP_METH, GMAP_KWMETH)
extern Py_ssize_t gmap_pysize(PyObject *);
extern PySequenceMethods gmap_pysequence;
extern PyMethodDef gmap_pymethods[];

/*----- Bytestrings -------------------------------------------------------*/

PyObject *bytestring_pywrap(const void *, size_t);
PyObject *bytestring_pywrapbuf(buf *);

/*----- Multiprecision arithmetic -----------------------------------------*/

typedef struct mp_pyobj {
  PyObject_HEAD
  mp *x;
} mp_pyobj;

extern PyTypeObject *mp_pytype;
extern PyTypeObject *gf_pytype;
#define MP_X(o) (((mp_pyobj *)(o))->x)
#define MP_PYCHECK(o) PyObject_TypeCheck((o), mp_pytype)
#define GF_PYCHECK(o) PyObject_TypeCheck((o), gf_pytype)

extern mp *mp_frompylong(PyObject *);
extern PyObject *mp_topylong(mp *);
extern mp *tomp(PyObject *);
extern mp *getmp(PyObject *);
extern int convmp(PyObject *, void *);
extern mp *getgf(PyObject *);
extern int convgf(PyObject *, void *);
extern PyObject *mp_pywrap(mp *);
extern PyObject *gf_pywrap(mp *);
extern long mphash(mp *);
extern mp *mp_frompyobject(PyObject *, int);
extern PyObject *mp_topystring(mp *, int,
			       const char *, const char *, const char *);
extern int mp_tolong_checked(mp *, long *, int);

/*----- Abstract fields ---------------------------------------------------*/

typedef struct field_pyobj {
  PyHeapTypeObject ty;
  field *f;
} field_pyobj;

extern PyTypeObject *field_pytype;
extern PyTypeObject *primefield_pytype;
extern PyTypeObject *niceprimefield_pytype;
extern PyTypeObject *binfield_pytype;
extern PyTypeObject *binpolyfield_pytype;
extern PyTypeObject *binnormfield_pytype;
#define FIELD_PYCHECK(o) PyObject_TypeCheck((o), field_pytype)
#define FIELD_F(o) (((field_pyobj *)(o))->f)
extern PyObject *field_pywrap(field *);
extern field *field_copy(field *);

typedef struct fe_pyobj {
  PyObject_HEAD
  field *f;
  mp *x;
} fe_pyobj;

extern PyTypeObject *fe_pytype;
#define FE_PYCHECK(o) PyObject_TypeCheck((o), fe_pytype)
#define FE_F(o) (((fe_pyobj *)(o))->f)
#define FE_FOBJ(o) ((PyObject *)(o)->ob_type)
#define FE_X(o) (((fe_pyobj *)(o))->x)
extern PyObject *fe_pywrap(PyObject *, mp *);

/*----- Elliptic curves ---------------------------------------------------*/

typedef struct eccurve_pyobj {
  PyHeapTypeObject ty;
  ec_curve *c;
  PyObject *fobj;
} eccurve_pyobj;

extern PyTypeObject *eccurve_pytype;
extern PyTypeObject *ecprimecurve_pytype;
extern PyTypeObject *ecprimeprojcurve_pytype;
extern PyTypeObject *ecbincurve_pytype;
extern PyTypeObject *ecbinprojcurve_pytype;
#define ECCURVE_PYCHECK(o) PyObject_TypeCheck((o), eccurve_pytype)
#define ECCURVE_C(o) (((eccurve_pyobj *)(o))->c)
#define ECCURVE_FOBJ(o) (((eccurve_pyobj *)(o))->fobj)
extern PyObject *eccurve_pywrap(PyObject *, ec_curve *);
extern ec_curve *eccurve_copy(ec_curve *);

typedef struct ecpt_pyobj {
  PyObject_HEAD
  ec_curve *c;
  ec p;
} ecpt_pyobj;

extern PyTypeObject *ecpt_pytype, *ecptcurve_pytype;
#define ECPT_PYCHECK(o) PyObject_TypeCheck((o), ecpt_pytype)
#define ECPTCURVE_PYCHECK(o) PyObject_TypeCheck((o), ecptcurve_pytype)
#define ECPT_C(o) (((ecpt_pyobj *)(o))->c)
#define ECPT_COBJ(o) ((PyObject *)(o)->ob_type)
#define ECPT_FOBJ(o) ECCURVE_FOBJ(ECPT_COBJ((o)))
#define ECPT_P(o) (&((ecpt_pyobj *)(o))->p)
extern PyObject *ecpt_pywrap(PyObject *, ec *);
extern PyObject *ecpt_pywrapout(void *, ec *);
extern int toecpt(ec_curve *, ec *, PyObject *);
extern int getecpt(ec_curve *, ec *, PyObject *);
extern void getecptout(ec *, PyObject *);
extern int convecpt(PyObject *, void *);

typedef struct ecinfo_pyobj {
  PyObject_HEAD
  ec_info ei;
  PyObject *cobj;
} ecinfo_pyobj;

extern PyTypeObject *ecinfo_pytype;
#define ECINFO_PYCHECK(o) PyObject_TypeCheck((o), ecinfo_pytype)
#define ECINFO_EI(o) (&((ecinfo_pyobj *)(o))->ei)
#define ECINFO_COBJ(o) (((ecinfo_pyobj *)(o))->cobj)
extern void ecinfo_copy(ec_info *, const ec_info *);
extern PyObject *ecinfo_pywrap(ec_info *);

/*----- Cyclic groups -----------------------------------------------------*/

typedef struct ge_pyobj {
  PyObject_HEAD
  ge *x;
  group *g;
} ge_pyobj;

extern PyTypeObject *ge_pytype;
#define GE_PYCHECK(o) PyObject_TypeCheck((o), ge_pytype)
#define GE_X(o) (((ge_pyobj *)(o))->x)
#define GE_G(o) (((ge_pyobj *)(o))->g)
#define GE_GOBJ(o) ((PyObject *)(group_pyobj *)(o)->ob_type)
extern PyObject *ge_pywrap(PyObject *, ge *);

typedef struct group_pyobj {
  PyHeapTypeObject ty;
  group *g;
} group_pyobj;

extern PyTypeObject *group_pytype;
#define GROUP_G(o) (((group_pyobj *)(o))->g)
extern PyObject *group_pywrap(group *);
extern group *group_copy(group *);

/*----- Random number generators ------------------------------------------*/

#define f_freeme 1u

typedef struct grand_pyobj {
  PyObject_HEAD
  unsigned f;
  grand *r;
} grand_pyobj;

extern PyTypeObject *grand_pytype;
extern PyObject *rand_pyobj;
#define GRAND_PYCHECK(o) PyObject_TypeCheck((o), grand_pytype)
#define GRAND_F(o) (((grand_pyobj *)(o))->f)
#define GRAND_R(o) (((grand_pyobj *)(o))->r)
extern PyObject *grand_pywrap(grand *, unsigned);
extern int convgrand(PyObject *, void *);

/*----- Symmetric cryptography --------------------------------------------*/

extern PyObject *keysz_pywrap(const octet *);

extern int convgccipher(PyObject *, void *);
extern PyObject *gccipher_pywrap(gccipher *);

typedef struct gchash_pyobj {
  PyHeapTypeObject ty;
  gchash *ch;
} gchash_pyobj;

extern PyTypeObject *gchash_pytype;
extern PyObject *sha_pyobj, *has160_pyobj;
#define GCHASH_PYCHECK(o) PyObject_TypeCheck((o), gchash_pytype)
#define GCHASH_CH(o) (((gchash_pyobj *)(o))->ch)
extern PyObject *ghash_pywrap(PyObject *, ghash *);
extern int convgchash(PyObject *, void *);
extern int convghash(PyObject *, void *);

extern int convgcmac(PyObject *, void *);

/*----- Key generation ----------------------------------------------------*/

typedef struct pfilt_pyobj {
  PyObject_HEAD
  pfilt f;
  int st;
} pfilt_pyobj;

extern PyTypeObject *pfilt_pytype;
#define PFILT_PYCHECK(o) PyObject_TypeCheck(o, pfilt_pytype)
#define PFILT_F(o) (&((pfilt_pyobj *)(o))->f)
#define PFILT_ST(o) (((pfilt_pyobj *)(o))->st)

typedef struct { pgen_proc *proc; void *ctx; } pgev;
#define PGEV_HEAD PyObject_HEAD pgev pg;

typedef struct pgev_pyobj {
  PGEV_HEAD
} pgev_pyobj;

extern PyTypeObject *pgev_pytype;
#define PGEV_PYCHECK(o) PyObject_TypeCheck(o, pgev_pytype)
#define PGEV_PG(o) (&((pgev_pyobj *)(o))->pg)

typedef struct pypgev {
  pgev ev;
  PyObject *obj;
  struct excinfo *exc;
} pypgev;

extern int convpgev(PyObject *, void *);
extern void droppgev(pypgev *);
extern void pgenerr(struct excinfo *exc);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
